-----------------------------------------------------------------------------------------------
-- Client Lua Script for UtilityCoordinator
-- Copyright (c) legowxelab2z8. All rights reserved
-- Slightly edited by Cargo Cult for API10 compatibility
-- Edited even more by Chill for taunt and utility coordination
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- A word about variable naming
-- Abilities and Spells in WildStar can get messy and confusing.
-- In an effort to reduce confusion for myself and anyone who is reading this I decided to define some names.
--
-- Abilities - Abilities are different than Spells.  Abilities are what make up your limited action set
--	ActionSetLib.GetCurrentActionSet() Can be used to get all of the character's active Abilitiy Ids.
--	AbilityBook.GetAbilitiesList() can be used to lookup info for that ability.
--		I walk the list looking for Ids returned by GetCurrentActionSet looking for a match.
--	AbilityBook.GetAbilityInfo() seems to always return null, and I can't find a carbine example.
--
-- Spells - Spells are different than Abilities.
--	Spells can be retrieved by tAbility.tTiers[tAbility.nCurrentTier].splObject
--		tAbility comes from GetAbilitiesList()
--	There are multiple Spells for each Ability.  Most Abiltilities have 9 Spells ("Base" rank and Tier 1-8)
--		Base Tier is [1] and Tier 1 is [2] in the tAbility.tTiers table.
--	GameLib.GetSpell() Can be used to retrieve a splObject for a nSpellId
--
-- SpellEffect - What I call a SpellEffect is the spell that is exposed by tEventArgs.splCallingSpell
--	For example the Esper spell Crush Tier 1 has a SpellId of 52050 but the ability that triggers
--	OnCombatLogCCState is 52059.  Both spells have the same GetName() and GetRank() values but
--	GetBaseSpellId() returns 26488 which is the spell CreatureGib.
--	Using a SpellEffect when you intend to use a Spell will cause issues with retrieving some spell
--	information such as GetCooldownTime() that returns 0 for the SpellEffect.
--
--	Not all Spells have a SpellEffect.  Some Spells use the same Spell for tEventArgs.splCallingSpell
--
-- A spell with a SpellEffect will have 18 different Spell Ids and 1 Ability Id!
-----------------------------------------------------------------------------------------------


require "Window"

-----------------------------------------------------------------------------------------------
-- UtilityCoordinator Module Definition
-----------------------------------------------------------------------------------------------
UtilityCoordinator = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local sVersion = "0.0.5"

local eCommMessageType =
  {
    LASChange = 1,
    CooldownChanged = 2,
    InformationRequest = 3,
  }
local filtermode = "none"
local bMove = false
--filtermodes
-- none
-- stuns
-- dgrids
-- bt
-- utility = bt,dgrid
-- taunts
--/it stuns
--

--This is a list of abilties that cause interrupts that we are interested in tracking
--Uses Ability.nId as key. (AbilityId not SpellId)
--Interrupts means that the ability will cause a CC effect on a target with 0 Interrupt Armor
--IARemovals is a table of Interrupt armor removed for each spell Tier.
--IARemovals[1] is Base Tier, IARemovals[2] is Tier1
local tInterruptSpellTable =
  {
    [19022] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Esper:Crush
    [19355] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Esper:Incapacitate
    [19029] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Esper:Shockwave
    [25635] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Engineer:Zap
    [34176] = {Interrupts = false, IARemovals = {1,1,1,1,1,1,1,1,1}}, -- Engineer:Obstruct Vision
    [34009] = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}}, 	-- Engineer:Blitz
    [20325] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Spellslinger:Gate
    [30160] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}},  -- Spellslinger:Arcane Shock
    [16454] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Spellslinger:Spatial Shift
    [26543] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Medic:Paralytic Surge
    [26529] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}},  -- Medic:Magnetic Lockdown
    [23173] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Stalker:Stagger
    [23587] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Stalker:False Retreat
    [23705] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Stalker:Collapse
    [38017] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Warrior:Kick
    [18547] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Warrior:Flash Bang
    [18363] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Warrior:Grapple


    [18569]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Warrior:Plasma Blast (Taunt)
    [18578]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Warrior:Atomic Surge (Taunt)

    [23892]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Stalker:Razor Storm (Taunt)
    [23704]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Stalker:Reaver (Taunt)

    [25819]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Engineer:Code Red (Taunt)
    [25472]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Engineer:Hyper Wave(Taunt)


    [35526]  = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}},   -- Warrior:Defense Grid
    [23829]  = {Interrupts = true, IARemovals = {0,1,2,3,4,5,6,7,8}},   -- Stalker:Bloodthirst


    -- use findspellid for this id
  }
local tTableStuns =
  {
    [19022] = true,
    [19355] = true,
    [19029] = true,
    [25635] = true,
    [34176] = true,
    [34009] = true,
    [20325] = true,
    [30160] = true,
    [16454] = true,
    [26543] = true,
    [23173] = true,
    [23587] = true,
    [23705] = true,
    [38017] = true,
    [18547] = true,
    [18363] = true,
    [26529] = true,
}
local tTableTaunts =
  {
    [18569]  = true, 
    [18578]  = true,   

    [23892]  = true,   
    [23704]  = true,  

    [25819]  = true, 
    [25472]  = true, 
}

local tTableUtil =
  {
    [23829]  = true,
    [35526]  = true, 
}

local tTabledgridandtaunt =
  {
    [35526]  = true,

    [18569]  = true, 
    [18578]  = true,   

    [23892]  = true,   
    [23704]  = true,  

    [25819]  = true, 
    [25472]  = true, 
}

local tTablebtandtaunt =
  {
    [23829]  = true,

    [18569]  = true, 
    [18578]  = true,   

    [23892]  = true,   
    [23704]  = true,  

    [25819]  = true, 
    [25472]  = true, 
}
--Used for Building SpellEffect Info for BuildSpellTable()
local tSpellNames =
  {
    ["Crush"] = true,
    ["Incapacitate"] = true,
    ["Shockwave"] = true,
    ["Zap"] = true,
    ["Obstruct Vision"] = true,
    ["Gate"] = true,
    ["Arcane Shock"] = true,
    ["Spatial Shift"] = true,
    ["Paralytic Surge"] = true,
    ["Magnetic Lockdown"] = true,
    ["Stagger"] = true,
    ["False Retreat"] = true,
    ["Collapse"] = true,
    ["Kick"] = true,
    ["Flash Bang"] = true,
    ["Grapple"] = true,
    ["Blitz"] = true,

    ["Plasma Blast"] = true,
    ["Code Red"] = true,

    ["Hyper Wave"] = true,
    ["Reaver"] = true,

    ["Razor Storm"] = true,
    ["Atomic Surge"] = true,

    ["Bloodthirst"] = true,
    ["Defense Grid"] = true,

  }

local tBlackWhiteList = {}

-----------------------------------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------------------------------

local Debug = false --Debugging flag for DebugPrint()

local bSendInfoRequest = false

--Used to store the last time the interrupt was cast
--Also used so we dont show wda message more than once for aoe interrupts
--[Player][nSpellId] = {timeLastCast, fCooldown}
local tLastCast = {}

local tSavedData --Used to store save data. Needed if OnDocLoaded hasn't happened yet.
local bDocLoaded = false -- Used to track if OnDocLoaded has been called.
local bRestored = false -- Used to track if OnRestore has been called.

local tSpellConversions = {} --Used for storing SpellEffect Info for BuildSpellTable()

local LastUpdateSent = nil  --Used for tracking the last time we sent a update comm message so requests dont cause us to spam.

local LastLeader = nil


function UtilityCoordinator:OnSave(eType)
  if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
  local tSaved = {
    tsavedFiltermode = filtermode,
    tGroupFrameData = {
      AnchorOffsets = {self.wndGroupFrame:GetAnchorOffsets()},
    }
  }
  return tSaved
end

-----------------------------------------------------------------------------------------------
-- OnRestore is not called the first time the addon is loaded.
-- Can be called before or after OnDocLoaded.
-----------------------------------------------------------------------------------------------
function UtilityCoordinator:OnRestore(eType, tRestoreData)
  tSavedData = tRestoreData
  if tSavedData.tsavedFiltermode ~= nil then
    filtermode = tSavedData.tsavedFiltermode
  end

  bRestored = true
  if bDocLoaded then --Frames are created.
    self:RestoreGroupFrame()
  end
end

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function UtilityCoordinator:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self

  -- initialize variables here
  self.tCurrentAbilities = {}
  self.tInterruptAbilities = {} --Used for tracking your own interrupt abilities.
  self.tCooldownTimers = {}
  self.PlayerFrames = {}
  self.CommChannel = nil
  self.CommPlayers = {}
  return o
end

function UtilityCoordinator:Init()
  LibJSON = Apollo.GetPackage("Lib:dkJSON-2.5").tPackage

  local bHasConfigureFunction = false
  local strConfigureButtonText = ""
  local tDependencies = {
    -- "UnitOrPackageName",
    }
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end


-----------------------------------------------------------------------------------------------
-- UtilityCoordinator OnLoad
-----------------------------------------------------------------------------------------------
function UtilityCoordinator:OnLoad()
  -- load our form file
  self.xmlDoc = XmlDoc.CreateFromFile("UtilityCoordinator.xml")
  self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end



-----------------------------------------------------------------------------------------------
-- UtilityCoordinator OnDocLoaded
-----------------------------------------------------------------------------------------------
function UtilityCoordinator:OnDocLoaded()

  if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
    self.wndMain = Apollo.LoadForm(self.xmlDoc, "UtilityCoordinatorForm", nil, self)
    if self.wndMain == nil then
      Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
      return
    end

    bDocLoaded = true

    self.wndMain:Show(false, true)

    self.wndGroupFrame = Apollo.LoadForm(self.xmlDoc, "GroupFrame", nil, self)
    self.wndGroupFrame:Show(true, true)

    if bRestored then --Saved data has already been loaded
      self:RestoreGroupFrame()
    end

    --add the player  (Bugs when first loading since addon loads before GetPlayerUnit will work)
    --self:AddPlayerFrame(GameLib.GetPlayerUnit():GetName())

    -- if the xmlDoc is no longer needed, you should set it to nil
    -- self.xmlDoc = nil

    -- Register handlers for events, slash commands and timer, etc.
    Apollo.RegisterEventHandler("CombatLogCCState", "OnCombatLogCCState", self)
    Apollo.RegisterEventHandler("BuffAdded", "OnBuffAdded", self)

    Apollo.RegisterSlashCommand("uc", "OnUtilityCoordinatorOn", self)
    Apollo.RegisterEventHandler("AbilityBookChange", "LASChanged", self)

    Apollo.RegisterEventHandler("Group_Updated", "OnGroupUpdated", self)
    Apollo.RegisterEventHandler("Group_Remove",	"OnGroupRemove", self)				-- ( name, reason )
    Apollo.RegisterEventHandler("Group_Add", "OnGroupAdd", self)					-- ( name )
    Apollo.RegisterEventHandler("Group_Left", "OnGroupLeft", self)				-- ( reason )


    Apollo.RegisterTimerHandler("BroadcastTimer", "NextFrame", self)
    Apollo.CreateTimer("BroadcastTimer", 0.7, true)
    Apollo.StartTimer("BroadcastTimer")

    -- Do additional Addon initialization here
  end
end

-----------------------------------------------------------------------------------------------
-- UtilityCoordinator Functions
-----------------------------------------------------------------------------------------------
function UtilityCoordinator:OnBuffAdded(caster, tbuff)
  if tbuff.splEffect:GetName() == "Bloodthirst" then
    if tbuff.splEffect:GetMaximumRange() == 0 then
      local dura = tbuff.fTimeRemaining
      local strCaster = caster:GetName()
      if dura == 4 then -- base
         fakeID = 39497
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 5 then -- t1
         fakeID = 59515
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 6 then -- t2
         fakeID = 59516
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 7 then -- t3
         fakeID = 59517
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 8 then -- t4
         fakeID = 59518
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 9 then -- t5
         fakeID = 59519
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 10 then-- t6
         fakeID = 59520
         fakespl = GameLib.GetSpell(fakeID)
      elseif dura == 11 then-- t7
         fakeID = 59521
         fakespl = GameLib.GetSpell(fakeID)
      else --t8
         fakeID = 59522
         fakespl = GameLib.GetSpell(fakeID)
      end
      if self.CommPlayers[strCaster] then return end
      self:TrackSpell(caster:GetName(), fakespl, caster)

    end
  end
end

function UtilityCoordinator:OnCombatLogCCState(tEventArgs)
  if not tEventArgs then return end
  if not tEventArgs.unitTarget then return end
  if not tEventArgs.unitCaster then return end
  if tEventArgs.bRemoved then return end -- indicates that the cc effect ended.


  --Event_FireGenericEvent("SendVarToRover", "LogCCState tEventArgs", tEventArgs, 0)

  local unitCaster = (tEventArgs.unitCasterOwner or tEventArgs.unitCaster)
  local strCaster = unitCaster:GetName()
  local strTarget = tEventArgs.unitTarget:GetName()
  local unitPlayer = GameLib.GetPlayerUnit()
  local splSpell = tEventArgs.splCallingSpell
  local nSpellId = splSpell:GetId()
  if not (unitCaster == GameLib.GetPlayerUnit() or unitCaster:IsInYourGroup()) then return end
  if self.CommPlayers[strCaster] then return end --This player has the addon, let them report instead of attempting to track

  --self:DebugPrint(strCaster .. " " .. splSpell:GetName() .. " " .. nSpellId .. " " .. splSpell:GetBaseSpellId())

  --The spell that shows up isn't always the same as the spell that is cast, so convert to spell cast if needed.
  if self.tSpellEffectIdConversion[nSpellId] then
    --self:DebugPrint("Converting ID " .. nSpellId .. " to " .. self.tSpellEffectIdConversion[nSpellId])
    nSpellId = self.tSpellEffectIdConversion[nSpellId]
    splSpell = GameLib.GetSpell(nSpellId)
  end

  local stunned = tEventArgs.unitTarget:GetCCStateTimeRemaining(tEventArgs.eState) > 0
  if stunned then
  --self:DebugPrint(strTarget .. " stunned.")
  end

  if self:checkifvalid(splSpell:GetBaseSpellId()) == false then return end
  if tInterruptSpellTable[splSpell:GetBaseSpellId()] then
    --self:DebugPrint("Tracking " .. splSpell:GetName())
    self:TrackSpell(unitCaster:GetName(), splSpell, unitCaster)
  else
  --[[

		tInterruptSpellTable[splSpell:GetBaseSpellId()] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}

		tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()] = tEventArgs.nInterruptArmorHit

		self:TrackSpell(unitCaster:GetName(), splSpell)

		self:DebugPrint(strCaster 

			.. " " .. splSpell:GetName() 

			.. " Id " .. nSpellId

			.. " Base Id " .. splSpell:GetBaseSpellId())

		]]
  end
end

--When AbilityBookChange event fires AbilityBook.GetAbilitiesList() doesn't always return the new list.
--Set a timer to give the game a chance to update the LAS before we grab it.
function UtilityCoordinator:LASChanged()
  --self:DebugPrint("LASChanged")
  self.RebuildLASTimer = ApolloTimer.Create(1, false, "RebuildLAS", self)
end

function UtilityCoordinator:RebuildLAS()
  if GameLib.GetPlayerUnit() == nil then self:LASChanged() return end --Set the timer again if not loaded yet.
  self:DebugPrint("RebuildLAS")
  local unitPlayer = GameLib.GetPlayerUnit()
  local tActionSetAbilties = ActionSetLib.GetCurrentActionSet()
  if not tActionSetAbilties then return end
  local tAllAbilities = AbilityBook.GetAbilitiesList()
  local tAbiltyIndexes = {}
  self.tInterruptAbilities = {}

  for indx = 1, 8 do
    if tActionSetAbilties[indx] > 0 then
      tAbiltyIndexes[tActionSetAbilties[indx]] = indx
    end
  end
  --self.wndMain:FindChild("EditBox"):SetText("")
  local strLASText = ""
  self.tCurrentAbilities = {}
  for i, tAbility in pairs(tAllAbilities) do
    if tAbility.nId and tAbiltyIndexes[tAbility.nId] then
      self.tCurrentAbilities[tAbiltyIndexes[tAbility.nId]] = tAbility
      strLASText = strLASText .. tAbility.strName .. " " .. tAbility.tTiers[tAbility.nCurrentTier].splObject:GetId() .. "\n"
    end
  end
  --self.wndMain:FindChild("EditBox"):SetText(strLASText)

  for i, tAbility in pairs(self.tCurrentAbilities) do
    --self:DebugPrint(tAbility.strName .. " " .. tAbility.nId)
    if tInterruptSpellTable[tAbility.nId] then
      local nCooldown = tAbility.tTiers[tAbility.nCurrentTier].splObject:GetCooldownTime()
      local strCooldown = string.format("%.2f", nCooldown)
      local nSpellId = tAbility.tTiers[tAbility.nCurrentTier].splObject:GetId()
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " Tier: " .. tAbility.nCurrentTier
        .. " Cooldown: " .. strCooldown
        .. " Removes: " .. tInterruptSpellTable[tAbility.nId].IARemovals[tAbility.nCurrentTier])

      self.tInterruptAbilities[tAbility.nId] = tAbility
      --self.tCooldownTimers[tAbility] = 0
      --self:AddPlayerSpell(unitPlayer:GetName(), nSpellId)
    end
  end

  self:SendCurrentInterrupts()
end

function UtilityCoordinator:NextFrame()

  if bSendInfoRequest then
    local bSuccess = self.CommChannel:SendMessage(LibJSON.encode({eType = eCommMessageType.InformationRequest}))
    if bSuccess then
      bSendInfoRequest = false
      self:RebuildLAS()
    end
  end

  --Get your own cooldowns
  for iAbilityId, tAbility in pairs(self.tInterruptAbilities) do
    local splCurrentSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
    local nSpellId = splCurrentSpell:GetId()
    local fCooldown = splCurrentSpell:GetCooldownRemaining()
    local fOldCooldown = self.tCooldownTimers[tAbility] or 0
    if fOldCooldown < fCooldown then --Cooldown increased
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " used. Cooldown: " .. fCooldown .. ".")
    self:SendCooldownInformation(nSpellId, fCooldown)
    elseif fOldCooldown - fCooldown > 1 then -- Cooldown changed a lot, report it again
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " Cooldown changed. New Cooldown: " .. fCooldown .. ".")
    self:SendCooldownInformation(nSpellId, fCooldown)
    end
    self.tCooldownTimers[tAbility] = fCooldown
  end

  -- Check if someone of the group is dead.
  local n = GroupLib.GetMemberCount()
  if n == 0 then
    return
  else
    for i=1, n do
      local info = GroupLib.GetGroupMember(i)
      if not info then return end
      local isDead = info.nHealth == 0 and info.nHealthMax ~= 0
      self:SetPlayerDisabled(info.strCharacterName, isDead)
    end
  end

  --Draw cooldown timers
  for strPlayerName,tSpells in pairs(tLastCast) do
    if self.PlayerFrames[strPlayerName] then
      local wndPlayerFrame = self.PlayerFrames[strPlayerName]
      for nSpellId,CooldownInfo in pairs(tSpells) do
        --self:DebugPrint(nSpellId)
        local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
        if wndSpellFrame then
          --for k,v in pairs(CooldownInfo) do self:DebugPrint(k .. " " .. v) end
          local nCooldown = CooldownInfo.nCooldown - Time.SecondsElapsed(CooldownInfo.timer)
          if nCooldown > 0.9 then
            local strCooldown
            strCooldown = string.format("%d", nCooldown)
            wndSpellFrame:SetText(strCooldown)
          else
            wndSpellFrame:SetText("")
            tLastCast[strPlayerName][nSpellId] = nil
            wndSpellFrame:FindChild("Overlay"):SetOpacity(0,1)
          end
        end
      end
    end
  end
end

function UtilityCoordinator:SetPlayerDisabled(playerName, disabled)
  if self.PlayerFrames[playerName] == nil then return end
  if disabled then
    self.PlayerFrames[playerName]:FindChild("Name"):SetTextColor(ApolloColor.new("red"))
  else
      if self.CommPlayers[playerName] ~= nil then
        self.PlayerFrames[playerName]:FindChild("Name"):SetTextColor(ApolloColor.new("green"))
      else
        self.PlayerFrames[playerName]:FindChild("Name"):SetTextColor(ApolloColor.new("white"))
      end
  end

end

function UtilityCoordinator:TrackSpell(strCaster, splSpell, uCaster)
  local nCooldown = splSpell:GetCooldownTime()
  local nSpellId = splSpell:GetId()
  if not tLastCast[strCaster] then tLastCast[strCaster] = {} end
  if not tLastCast[strCaster][nSpellId] then tLastCast[strCaster][nSpellId] = {} end
  tLastCast[strCaster][nSpellId] = {timer = Time.Now(),nCooldown = nCooldown}

  self:AddPlayerSpell(strCaster, nSpellId, uCaster)
  local wndPlayerFrame = self.PlayerFrames[strCaster]
  if not wndPlayerFrame then self:DebugPrint("No Player frame" .. strCaster) return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
  if not wndSpellFrame then self:DebugPrint("No spell frame" .. strCaster .. " " .. nSpellId) return end
  wndSpellFrame:FindChild("Overlay"):SetOpacity(.7,0)
end


function UtilityCoordinator:TrackReportedSpell(sPlayerName, nSpellId, fCooldown)
  if not tLastCast[sPlayerName] then tLastCast[sPlayerName] = {} end
  if not tLastCast[sPlayerName][nSpellId] then tLastCast[sPlayerName][nSpellId] = {} end
  tLastCast[sPlayerName][nSpellId] = {timer = Time.Now(),nCooldown = fCooldown}
  self:AddPlayerSpell(sPlayerName, nSpellId)
  local wndPlayerFrame = self.PlayerFrames[sPlayerName]
  if not wndPlayerFrame then self:DebugPrint("No Player frame" .. sPlayerName) return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
  if not wndSpellFrame then self:DebugPrint("No spell frame" .. sPlayerName .. " " .. nSpellId) return end
  wndSpellFrame:FindChild("Overlay"):SetOpacity(.7,0)
end


function UtilityCoordinator:OnGroupUpdated()
  for idx = 1, GroupLib.GetMemberCount() do
    local tMemberInfo = GroupLib.GetGroupMember(idx)
   -- if tMemberInfo ~= nil then
    --  self:AddPlayerFrame(tMemberInfo.strCharacterName)
   -- end
  end
  self.wndGroupFrame:ArrangeChildrenVert(0, sortByName)
  self:GetGroupLeader()
end

function UtilityCoordinator:OnGroupAdd(sName)
  --if self:AddPlayerFrame(sName) then
  --  self:DebugPrint("OnGroupAdd - " .. sName .. " added.")
 -- end
end

--Fires when a player leaves the group
function UtilityCoordinator:OnGroupRemove(sName, eReason)
  self:DebugPrint("OnGroupRemove")
  self:RemovePlayerFrame(sName)
  self.wndGroupFrame:ArrangeChildrenVert(0, sortByName)
end

--Fires when the player leaves a group
function UtilityCoordinator:OnGroupLeft(eReason)
  self:DebugPrint("OnGroupLeft")
  local sMyName = GameLib.GetPlayerUnit():GetName()
  for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
    self:RemovePlayerFrame(sPlayerName)
  end
  self.CommChannel = nil
  LastLeader = nil

end

function UtilityCoordinator:GetGroupLeader()
  if not GroupLib.InGroup() then LastLeader = nil return end
  local sLeader = nil
  for idx = 1, GroupLib.GetMemberCount() do
    local tMemberInfo = GroupLib.GetGroupMember(idx)
    if tMemberInfo ~= nil then
      if tMemberInfo.bIsLeader then sLeader = tMemberInfo.strCharacterName end
    end
  end
  if sLeader ~= LastLeader then -- Leader changed
    self:GroupLeaderChanged(sLeader)
  end
  LastLeader = sLeader
  return sLeader
end

function UtilityCoordinator:GroupLeaderChanged(sLeaderName)
  self:DebugPrint("GroupLeaderChanged")
  self:DebugPrint("Joinning Channel")
  self.CommChannel = ICCommLib.JoinChannel("UtilityCoordinator" .. sLeaderName, ICCommLib.CodeEnumICCommChannelType.Group)
  self.CommChannel:SetReceivedMessageFunction("OnChannelMessage", self)
  bSendInfoRequest = true
end

function UtilityCoordinator:OnChannelMessage(channel, strMsg, sSenderName)
  self:DebugPrint("OnChannelMessage " .. strMsg)
  local tMsg = LibJSON.decode(strMsg)

  if not tMsg then return end

  Event_FireGenericEvent("SendVarToRover", "ChannelMessage", tMsg, 0)
  --if self.PlayerFrames[sSenderName] then --They are in our party, process message.
    --self:DebugPrint("Message from " .. sSenderName)
    

    if tMsg.eType == eCommMessageType.LASChange then
      --Print("got las")
      self:DebugPrint(sSenderName .. " LASChange Message")
      self:RemoveAllPlayerSpells(sSenderName)


      for idx, nSpellId in ipairs(tMsg) do
      	local splSpell = GameLib.GetSpell(nSpellId)
      	self:DebugPrint(splSpell .. " adding to playerspells...")
        self:AddPlayerSpell(sSenderName, nSpellId)
        self:DebugPrint(nSpellId .. " added from LAS MSG")
      end
    elseif tMsg.eType == eCommMessageType.CooldownChanged then
      self:DebugPrint("Cooldown Info Recieved " .. sSenderName)

      local splSpell = GameLib.GetSpell(tMsg.tCooldownInfo.nSpellId)
      if tInterruptSpellTable[splSpell:GetBaseSpellId()] then
        self:TrackSpell(sSenderName, splSpell)
      end

      self:TrackReportedSpell(sSenderName, tMsg.tCooldownInfo.nSpellId, tMsg.tCooldownInfo.fCooldown)
    elseif tMsg.eType == eCommMessageType.InformationRequest then
      self:DebugPrint("InformationRequest recieved " .. sSenderName)
      self:RebuildLAS() --Triggers a send message
    end
    
  if self.PlayerFrames[sSenderName] ~= nil then
    self.CommPlayers[sSenderName] = true
    self.PlayerFrames[sSenderName]:FindChild("Name"):SetTextColor(ApolloColor.new("green"))
  end
end
  
function UtilityCoordinator:SendCurrentInterrupts()
  if not GroupLib.InGroup() or not self.CommChannel then return end
  self:DebugPrint("Sending Interrupts")
  local tInterruptSpells = {eType = eCommMessageType.LASChange}
  for iAbilityId, tAbility in pairs(self.tInterruptAbilities) do
    local splCurrentSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
    local nSpellId = splCurrentSpell:GetId()
    if nSpellId then
      table.insert(tInterruptSpells, nSpellId)
    end
  end
  local strMessage = LibJSON.encode(tInterruptSpells)
  self.CommChannel:SendMessage(strMessage)
  self:OnChannelMessage(self.CommChannel, strMessage, GameLib.GetPlayerUnit():GetName())
  Event_FireGenericEvent("SendVarToRover", "tInterruptSpells", tInterruptSpells, 0)
  self.CDSendTimer = ApolloTimer.Create(1, false, "SendCurrentInterruptsCD", self)
end

function UtilityCoordinator:SendCurrentInterruptsCD()
  if not GroupLib.InGroup() or not self.CommChannel then return end
  self:DebugPrint("Sending Interrupts CD fixed")
  local tInterruptSpells = {eType = eCommMessageType.LASChange}
  for iAbilityId, tAbility in pairs(self.tInterruptAbilities) do
    local splCurrentSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
    local fCooldown = splCurrentSpell:GetCooldownRemaining()
    local nSpellId = splCurrentSpell:GetId()
    if nSpellId then
      self:SendCooldownInformation(nSpellId, fCooldown)
      table.insert(tInterruptSpells, nSpellId)
    end
  end
  local strMessage = LibJSON.encode(tInterruptSpells)
  --self.CommChannel:SendMessage(strMessage)
  --self:OnChannelMessage(self.CommChannel, strMessage, GameLib.GetPlayerUnit():GetName())
  --Event_FireGenericEvent("SendVarToRover", "tInterruptSpells", tInterruptSpells, 0)
end

function UtilityCoordinator:SendCooldownInformation(nSpellId, fCooldown)
  if not nSpellId then return end
  if not fCooldown then return end
  if not GroupLib.InGroup() or not self.CommChannel then return end
  self:DebugPrint("Sending cooldown " .. nSpellId)
  local tCooldownMessage = {
    eType = eCommMessageType.CooldownChanged,
    tCooldownInfo = {nSpellId = nSpellId, fCooldown = fCooldown},
  }
  
  local strMessage = LibJSON.encode(tCooldownMessage)
  
  self.CommChannel:SendMessage(strMessage)
  self:OnChannelMessage(self.CommChannel, strMessage, GameLib.GetPlayerUnit():GetName())
end

--Builds the table of spell casts and effects by walking though all the spell ids
--Only works for english localization
--Finds more Effect Ids than are needed, but I don't know a pragmatic way to filter them yet.
--Since we only care about player casts erroneous effect ids (such as a boss attack called Crush duplicating the Esper spell)
--Hopefully won't cause issues.
function UtilityCoordinator:BuildSpellTable()
  for nSpellId = 1, 100000 do
    local splSpell = GameLib.GetSpell(nSpellId)
    if splSpell then
      local strName = splSpell:GetName()
      if tSpellNames[strName] then --Spell is an interrupt we want to track
        local nBaseSpell = splSpell:GetBaseSpellId()
        local nTier = splSpell:GetTier()
        if not tSpellConversions[strName] then tSpellConversions[strName] = {} end
        if not tSpellConversions[strName][nTier] then tSpellConversions[strName][nTier] = {} end

        if tInterruptSpellTable[nBaseSpell] then --Spell is the actual cast version of the spell not the effect
          tSpellConversions[strName][nTier].Cast = nSpellId
        else
          if not tSpellConversions[strName][nTier].Effect then tSpellConversions[strName][nTier].Effect = {} end
          table.insert(tSpellConversions[strName][nTier].Effect, nSpellId)
          --tSpellConversions[strName][nTier].Effect = nSpellId
        end
      end
    end
  end
  Event_FireGenericEvent("SendVarToRover", "tSpellConversions", tSpellConversions, 0)

  local strConversionTables = "local tSpellEffectIdConversion = \n{\n"
  for strSpellName,tConversionTable in pairs(tSpellConversions) do
    for nTier, tSpellInfo in pairs(tConversionTable) do
      local nCastId = tSpellInfo.Cast or 0
      if tSpellInfo.Effect and type(tSpellInfo.Effect) == "table" then
        for indx, nEffectId in pairs(tSpellInfo.Effect) do
          strConversionTables = strConversionTables .. "\t[" .. nEffectId .. "] = " .. nCastId .. ", --" .. strSpellName .. ":Tier" ..nTier-1 .."\n"
        end
      end
    end
  end
  --strConversionTables = strConversionTables .. "}"
  --self.wndMain:Invoke() -- show the window
  
  
  --self.wndMain:FindChild("EditBox"):SetText(strConversionTables)

end

-- on SlashCommand "/it"
function UtilityCoordinator:OnUtilityCoordinatorOn(strcmd, arg)

  if arg == "all" then
    filtermode = "none"
    Print("UtilityCoordinator mode: no filter")

  elseif arg == "stuns" then
    filtermode = "stuns"
    Print("UtilityCoordinator mode: only stuns")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end

  elseif arg == "dgrids" then
    filtermode = "dgrids"
    Print("UtilityCoordinator mode: only dgrids")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end

  elseif arg == "bt" then
    filtermode = "bt"
    Print("UtilityCoordinator mode: only blood thirst")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end

  elseif arg == "utility" then
    filtermode = "utility"
    Print("UtilityCoordinator mode: blood thirst and dgrids")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end

  elseif arg == "dgridsandtaunts" then
    filtermode = "dgridsandtaunts"
    Print("UtilityCoordinator mode: dgrids and taunts")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end

  elseif arg == "taunts" then
    filtermode = "taunts"
    Print("UtilityCoordinator mode: only taunts")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end
  elseif arg == "off" then
    filtermode = "off"
    Print("UtilityCoordinator mode: off")

    for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
    end
    self.wndGroupFrame:Show(false, true)
  elseif arg == "current" then
    Print("UtilityCoordinator mode: " .. filtermode)
  elseif arg == "on" then
    Print("UtilityCoordinator mode: on")
    self.wndGroupFrame:Show(true, true)
    filtermode = "none"
  elseif arg == "move" then
    if bMove == false then
       Print("UtilityCoordinator mode: moveable mode activated")
       bMove = true
       self.wndGroupFrame:SetStyle("Moveable", true)

    else
       Print("UtilityCoordinator mode: moveable mode deactivated")
       bMove = false
       self.wndGroupFrame:SetStyle("Moveable", false)
    end


  else

  self.wndMain:Invoke()

  if filtermode == "dgrids" then
    self.wndMain:FindChild("ButtonDgrid"):SetCheck(true)
  elseif filtermode == "none" then
    self.wndMain:FindChild("ButtonEverything"):SetCheck(true)
  elseif filtermode == "stuns" then
    self.wndMain:FindChild("ButtonStuns"):SetCheck(true)
  elseif filtermode == "bt" then
    self.wndMain:FindChild("ButtonBt"):SetCheck(true)
  elseif filtermode == "utility" then
    self.wndMain:FindChild("ButtonDgridBt"):SetCheck(true)
  elseif filtermode == "taunts" then
    self.wndMain:FindChild("ButtonTaunts"):SetCheck(true)
  elseif filtermode == "dgridsandtaunts" then
    self.wndMain:FindChild("ButtonDgridTaunts"):SetCheck(true)
  elseif filtermode == "btandtaunts" then
    self.wndMain:FindChild("ButtonBtTaunt"):SetCheck(true)
  elseif filtermode == "off" then
    self.wndMain:FindChild("ButtonNothing"):SetCheck(true)
  else
    self.wndMain:FindChild("ButtonNothing"):SetCheck(true)
  end

  if bMove == false then
    self.wndMain:FindChild("ButtonLock"):SetCheck(false)
  else
    self.wndMain:FindChild("ButtonLock"):SetCheck(true)
  end

  if 1 == 1 then
    return
  end

    self.wndMain:Invoke() -- show the window
     for nSpellId = 1, 100000 do
      local splSpell = GameLib.GetSpell(nSpellId)
      if splSpell then
        local strName = splSpell:GetName()
        if tSpellNames[strName] then --Spell is an interrupt we want to track
          local nBaseSpell = splSpell:GetBaseSpellId()
          local nTier = splSpell:GetTier()

          if not tSpellConversions[strName] then tSpellConversions[strName] = {} end
          if not tSpellConversions[strName][nTier] then tSpellConversions[strName][nTier] = {} end

          if tInterruptSpellTable[nBaseSpell] then --Spell is the actual cast version of the spell not the effect
            tSpellConversions[strName][nTier].Cast = nSpellId
          else
            if not tSpellConversions[strName][nTier].Effect then tSpellConversions[strName][nTier].Effect = {} end
            table.insert(tSpellConversions[strName][nTier].Effect, nSpellId)
            --tSpellConversions[strName][nTier].Effect = nSpellId
          end
        end
      end
    end
    Event_FireGenericEvent("SendVarToRover", "tSpellConversions", tSpellConversions, 0)

    local strConversionTables = "local tSpellEffectIdConversion = \n{\n"
    for strSpellName,tConversionTable in pairs(tSpellConversions) do
      for nTier, tSpellInfo in pairs(tConversionTable) do
        local nCastId = tSpellInfo.Cast or 0
        if tSpellInfo.Effect and type(tSpellInfo.Effect) == "table" then
          for indx, nEffectId in pairs(tSpellInfo.Effect) do
            strConversionTables = strConversionTables .. "\t[" .. nEffectId .. "] = " .. nCastId .. ", --" .. strSpellName .. ":Tier" ..nTier-1 .."\n"
          end
        end
      end
    end
    strConversionTables = strConversionTables .. "}"
    
     self.wndMain:FindChild("EditBox"):SetText(strConversionTables)
    end
end

function UtilityCoordinator:RadioChecked(wndHandler, wndControl, eMouseButton)
  local buttonName = wndControl:GetName()
  if buttonName == "ButtonDgrid" then
    filtermode = "dgrids"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonEverything" then
    filtermode = "none"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonStuns" then
    filtermode = "stuns"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonBt" then
    filtermode = "bt"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonDgridBt" then
    filtermode = "utility"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonTaunts" then
    filtermode = "taunts"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonDgridTaunts" then
    filtermode = "dgridsandtaunts"
    self.wndGroupFrame:Show(true, true)
  elseif buttonName == "ButtonNothing" then
    filtermode = "off"
    self.wndGroupFrame:Show(false, true)
  elseif buttonName == "ButtonBtTaunt" then
    filtermode = "btandtaunts"
    self.wndGroupFrame:Show(true, true)
  else
    filtermode = "off"
    self.wndGroupFrame:Show(false, true)
  end
  for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
     self:RemovePlayerFrame(sPlayerName)
  end
end
function UtilityCoordinator:LockChecked(wndHandler, wndControl, eMouseButton)
  bMove = false
  self.wndGroupFrame:SetStyle("Moveable", false)

end
function UtilityCoordinator:LockUnchecked(wndHandler, wndControl, eMouseButton)
  bMove = true
  self.wndGroupFrame:SetStyle("Moveable", true)

end

function UtilityCoordinator:DebugPrint(strDebugString)
  if Debug then
    Print(strDebugString)
  end
  --ChatSystemLib.Command("/p " .. strDebugString)
end


-----------------------------------------------------------------------------------------------
-- UtilityCoordinatorForm Functions
-----------------------------------------------------------------------------------------------

function UtilityCoordinator:AddPlayerFrame(strPlayerName)
  if not strPlayerName then return false end
  if self.PlayerFrames[strPlayerName] then return end

  local wndPlayerFrame = Apollo.LoadForm(self.xmlDoc, "PlayerForm", self.wndGroupFrame, self)

  wndPlayerFrame:SetName(strPlayerName)
  wndPlayerFrame:FindChild("Name"):SetText(string.sub(strPlayerName,1,7))

  self.PlayerFrames[strPlayerName] = wndPlayerFrame

  --self.wndGroupFrame:ArrangeChildrenHorz()
  --self.wndGroupFrame:ArrangeChildrenHorz(0, sortByName)
  self.wndGroupFrame:ArrangeChildrenVert(0, sortByName)
  self:ResizeGroupFrame()
  return true
end

function UtilityCoordinator:RemovePlayerFrame(strPlayerName)
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  wndPlayerFrame:Destroy()
  self.PlayerFrames[strPlayerName] = nil
  self.CommPlayers[strPlayerName] = nil
  self:ResizeGroupFrame()
end
function UtilityCoordinator:checkifvalid(nBaseSpell)
  if filtermode == "none" then
    return true
  elseif filtermode == "stuns" then
    if tTableStuns[nBaseSpell] ~= nil then
      return true
    else
      return false
    end
  elseif filtermode == "dgrids" then
     if nBaseSpell == 35526 then
      return true
    else
      return false
    end

  elseif filtermode == "bt" then
     if nBaseSpell == 23829 then
      return true
    else
      return false
    end

  elseif filtermode == "utility" then
     if tTableUtil[nBaseSpell] ~= nil then
      return true
    else
      return false
    end

  elseif filtermode == "dgridsandtaunts" then
      if tTabledgridandtaunt[nBaseSpell] ~= nil then
        return true
      else
        return false
      end

  elseif filtermode == "btandtaunts" then
      if tTablebtandtaunt[nBaseSpell] ~= nil then
        return true
      else
        return false
      end

  elseif filtermode == "taunts" then
     if tTableTaunts[nBaseSpell] ~= nil then
      return true
    else
      return false
    end
  else

    return false
  end
end


function UtilityCoordinator:AddPlayerSpell(strPlayerName, nSpellId, uCaster)
  self:DebugPrint("AddPlayerSpell")
  if not nSpellId then return end
  --self:DebugPrint("s1")
  local splSpell = GameLib.GetSpell(nSpellId)
  if not splSpell then return end
  --self:DebugPrint("s2")
  local nBaseSpell = splSpell:GetBaseSpellId()

  if self:checkifvalid(nBaseSpell) == false then return end
  self:DebugPrint("after valid check")

  if not self.PlayerFrames[strPlayerName] then self:AddPlayerFrame(strPlayerName, uCaster) end
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]

  --Make/Update the spellframe
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId) --Try to get the frame if it already exists
  if wndSpellFrame == nil then
    wndSpellFrame = Apollo.LoadForm(self.xmlDoc, "CooldownIconForm", wndPlayerFrame, self)
  end
  wndSpellFrame:SetName(nSpellId)
  local strSpellIcon = splSpell:GetIcon()
  if not strSpellIcon then return end
  wndSpellFrame:SetSprite(strSpellIcon)
  wndSpellFrame:FindChild("Overlay"):SetOpacity(0,1)
  --self:DebugPrint(splSpell:GetBaseSpellId() .. " " .. splSpell:GetTier())
  local nTier = splSpell:GetTier()
  local sTier = ""
  if nTier == 1 then
    sTier = "Base"
  else
    sTier = "Tier: " .. tostring(nTier - 1)
  end

  wndSpellFrame:SetTooltip(splSpell:GetName() .. " " ..  sTier)

  --Add the Interrupt Armor Removal number
  local nIARemovals = "?"
  if tInterruptSpellTable[splSpell:GetBaseSpellId()] and
    tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()] then
    nIARemovals = tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()]
  end
  if not nIARemovals then nIARemovals = "?" end
  --self:DebugPrint("Removes " .. nIARemovals)
  wndSpellFrame:FindChild("IARemovalsForm"):SetText(nIARemovals)


  --Remove old spell ranks
  --Needed when tracking spells from other players when they don't have the addon to self report.
  local twndFrames = wndPlayerFrame:GetChildren()
  for indx, wndFrame in pairs(twndFrames) do
    local sName = wndFrame:GetName()
    local nFrameId = tonumber(sName)
    if nFrameId and nFrameId ~= nSpellId then
      local splFrameSpell = GameLib.GetSpell(nFrameId)
      if splFrameSpell then
        if splFrameSpell:GetBaseSpellId() == nBaseSpell then --Matchin base spells means duplicate (different tiers)
          self:RemovePlayerSpell(strPlayerName, nFrameId)
        end
      end
    end
  end

  --Update the height
  self:ResizePlayerFrame(wndPlayerFrame)

end

function UtilityCoordinator:RemovePlayerSpell(strPlayerName, nSpellId)
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId) --Try to get the frame if it already exists
  if not wndSpellFrame then return end
  wndSpellFrame:Destroy()
  self:ResizePlayerFrame(wndPlayerFrame)
end

function UtilityCoordinator:RemoveAllPlayerSpells(strPlayerName)
  self:DebugPrint(strPlayerName .. " deleting all player spells...")
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  self:DebugPrint(strPlayerName .. " found playerframe")
  for idx, wndSpellFrame in pairs(wndPlayerFrame:GetChildren()) do
    local nSpellId = tonumber(wndSpellFrame:GetName()) or 0
    if GameLib.GetSpell(nSpellId) then
     self:DebugPrint(nSpellId .. " deleting player spell")
      self:RemovePlayerSpell(strPlayerName, nSpellId)
    end
  end
  self:RemovePlayerFrame(strPlayerName)
  self:ResizePlayerFrame(wndPlayerFrame)
  self:DebugPrint(strPlayerName .. " resized the playerframe")
end

function UtilityCoordinator:ResizePlayerFrame(wndFrame)
  local left,top,right,bottom = wndFrame:GetAnchorOffsets()
  local width = 0
  local height = 0
  for k,v in pairs(wndFrame:GetChildren()) do
    height = math.max(height, v:GetHeight())
    width = width + v:GetWidth()
  end
  wndFrame:SetAnchorOffsets(left,top,left+width,top+height)
  --wndFrame:SetAnchorOffsets(left,top,right+width,top)

  --wndFrame:ArrangeChildrenVert(1)
  wndFrame:ArrangeChildrenHorz(1)
  self:ResizeGroupFrame()
end

function UtilityCoordinator:ResizeGroupFrame()
  local width = 0
  local height = 0
  for k,wndPlayer in pairs(self.wndGroupFrame:GetChildren()) do
    --width = width + wndPlayer:GetWidth()
    --height = math.max(height, wndPlayer:GetHeight())

    width = math.max(width, wndPlayer:GetWidth())
    height = height + wndPlayer:GetHeight()
  end
  local left,top,right,bottom = self.wndGroupFrame:GetAnchorOffsets()
  self.wndGroupFrame:SetAnchorOffsets(left,top,left+width,top+height)
end

function UtilityCoordinator:RestoreGroupFrame()
  if bRestored then --Cant restore data if the data hasn't loaded
    local tGroupFrameData = tSavedData.tGroupFrameData
    if not tGroupFrameData or type(tGroupFrameData) ~= "table" then return end
    if tGroupFrameData.AnchorOffsets then
      self.wndGroupFrame:SetAnchorOffsets(unpack(tSavedData.tGroupFrameData.AnchorOffsets))
    end
  end
end

-- when the OK button is clicked
function UtilityCoordinator:OnOK()
  self.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function UtilityCoordinator:OnCancel()
  self.wndMain:Close() -- hide the window
end

function UtilityCoordinator:ClickPlayer(wndHandler, wndControl, eMouseButton)
	if eMouseButton == GameLib.CodeEnumInputMouse.Right then
		local wndName = wndControl:FindChild("Name")
		if wndName then
			local sPlayerName = wndName:GetText()
			if sPlayerName then
				self:RemovePlayerFrame(sPlayerName)
				tBlackWhiteList[sPlayerName] = true
			end
		end



	end
end



function UtilityCoordinator:CooldownIconFormOnMouseButtonUp(wndHandler, wndControl, eMouseButton)
  if eMouseButton == GameLib.CodeEnumInputMouse.Right then
    --clunky and should probably be updated
    local nSpellId = tonumber(wndControl:GetName())
    if nSpellId then
      local sPlayerName = wndControl:GetParent():FindChild("Name"):GetText()
      --self:DebugPrint("Mouse Up " .. wndControl:GetName() .. " " .. sPlayerName)
      if sPlayerName then
        self:RemovePlayerSpell(sPlayerName, nSpellId)
      end
    end
  end
end

function sortByName(a, b)
  return a:GetName() < b:GetName()
end

function UtilityCoordinator:BuildSpellDebugEditBox()

end

-----------------------------------------------------------------------------------------------
-- UtilityCoordinator Instance
-----------------------------------------------------------------------------------------------
local UtilityCoordinatorInst = UtilityCoordinator:new()
UtilityCoordinatorInst:Init()
